#include <string>
#include <iterator>
#include <algorithm>
#include <functional>
#include <utility>
#include <limits>
#include <cstdio>
#include <cstdlib>

#define STRICT
#include <windows.h>
#include <shlwapi.h>
#include <tlhelp32.h>

#define NELEMS(array) (sizeof(array)/sizeof((array)[0]))

using namespace std;

typedef unsigned long ulong;
typedef unsigned int uint;
typedef unsigned short ushort;
typedef unsigned char uchar;

namespace{
    const char EXT_EXE[] = ".exe";
    const char EXT_DLL[] = ".dll";

    const int MY_PATH_MAX = 4096;

    char tolower_ascii(char c)
    {
        return 'A'<=c && c<='Z' ? c-'A'+'a' : c;
    }

    string str_lower_ascii(const string& s)
    {
        string lower;
        transform(s.begin(), s.end(), back_inserter(lower), tolower_ascii);
        return lower;
    }

    struct ChCmpIAscii : public binary_function<char, char, int>{
        int operator()(char c1, char c2) const
        {
            char lc1 = tolower_ascii(c1);
            char lc2 = tolower_ascii(c2);
            if(lc1 < lc2) return -1;
            if(lc1 > lc2) return 1;
            return 0;
        }
    };
    int str_cmpi_ascii_impl(const string& s1, const string& s2)
    {
        typedef pair<string::const_iterator, string::const_iterator> PSCI;
        PSCI p = mismatch(s1.begin(), s1.end(), s2.begin(), not2(ChCmpIAscii()));
        if(p.first == s1.end()){
            if(p.second == s2.end()) return 0;
            else return -1;
        }
        return ChCmpIAscii()(*p.first, *p.second);
    }
    int str_cmpi_ascii(const string& s1, const string& s2)
    {
        if(s1.size() <= s2.size())
            return str_cmpi_ascii_impl(s1, s2);
        else
            return -str_cmpi_ascii_impl(s2, s1);
    }

    bool endswith(const string& s, const string& suffix)
    {
        return s.rfind(suffix) + suffix.size() == s.size();
    }

    bool my_realpath(const string& from, string& to)
    {
        char buf[MY_PATH_MAX];
        if(!_fullpath(buf, from.c_str(), NELEMS(buf))) return false;
        to = buf;
        return true;
    }

    void error(const string& msg)
    {
        fputs(msg.c_str(), stderr);
        putc('\n', stderr);
        exit(1);
    }

    ulong exe2pid(const string& exe)
    {
        HANDLE hSnap = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
        if(hSnap == INVALID_HANDLE_VALUE) return 0;

        ulong pid = 0;
        PROCESSENTRY32 pe32 = { sizeof(PROCESSENTRY32) };
        for(BOOL more = Process32First(hSnap, &pe32); more; more = Process32Next(hSnap, &pe32)){
            if(str_cmpi_ascii(exe, pe32.szExeFile) == 0){
                pid = pe32.th32ProcessID;
                break;
            }
        }

        CloseHandle(hSnap);
        return pid;
    }

    bool inject(ulong pid, const string& dll)
    {
        bool ret = false;
        HANDLE hProc = 0;
        HMODULE hMod = 0;
        FARPROC loadLibrary = 0;
        char* rmem = 0;
        HANDLE hThread = 0;

        DWORD access = PROCESS_CREATE_THREAD | PROCESS_VM_OPERATION | PROCESS_VM_WRITE;
        if(!(hProc = OpenProcess(access, FALSE, pid))) goto FAIL;
        if(!(hMod = GetModuleHandle("kernel32.dll"))) goto FAIL;
        if(!(loadLibrary = GetProcAddress(hMod, "LoadLibraryA"))) goto FAIL;
        if(!(rmem = static_cast<char*>(VirtualAllocEx(hProc, 0, dll.size()+1,
                                                      MEM_COMMIT|MEM_RESERVE, PAGE_READWRITE)))) goto FAIL;
        SIZE_T nwritten;
        if(!WriteProcessMemory(hProc, rmem, dll.c_str(), dll.size()+1, &nwritten)) goto FAIL;
        if(nwritten != dll.size()+1) goto FAIL;
        if(!(hThread = CreateRemoteThread(hProc, 0, 0, reinterpret_cast<LPTHREAD_START_ROUTINE>(loadLibrary), rmem,
                                          0, 0))) goto FAIL;
        WaitForSingleObject(hThread, INFINITE);

        ret = true;
FAIL:
        if(hThread) CloseHandle(hThread);
        if(rmem) VirtualFreeEx(hProc, rmem, 0, MEM_RELEASE);
        if(hProc) CloseHandle(hProc);
        return ret;
    }

    void usage()
    {
        error("Usage: dllinject <PID/EXE> <DLL>");
    }
}

int main(int argc, char** argv)
{
    if(argc != 3) usage();
    ulong pid = strtoul(argv[1], 0, 0);
    if(!pid || pid == numeric_limits<ulong>::max()){
        if(!endswith(str_lower_ascii(argv[1]), EXT_EXE)) usage();
        if(!(pid = exe2pid(argv[1]))) error("No such process");
    }
    if(!endswith(str_lower_ascii(argv[2]), EXT_DLL)) usage();
    if(!PathFileExists(argv[2])) error("No such DLL");
    string dll_real;
    if(!my_realpath(argv[2], dll_real)) error("Invalid DLL path");

    printf("PID: %lu\n", pid);
    printf("DLL: %s\n", dll_real.c_str());

    if(!inject(pid, dll_real)) error("Injection failed");

    return 0;
}
